#extension GL_EXT_shader_texture_lod : enable
#extension GL_OES_standard_derivatives : enable

precision highp float;

// Input variables for WebGL 1.0
varying vec3 v_position;
varying vec4 v_uv;
varying vec3 v_normal;
varying vec3 v_tangentuws;
varying vec3 v_tangentvws;
varying vec3 v_view;

/// DIRECTIONAL LIGHT SHADOW UNIFORMS AND VARYINGS
varying vec4 v_PositionFromDirLight;
uniform sampler2D _DirectionalShadowSampler;
uniform float shadowDarkness;

/// SPOT LIGHT SHADOW UNIFORMS AND VARYINGS
varying vec4 v_PositionFromSpotLight;
varying float v_DepthMetric;
uniform sampler2D _SpotLightShadowSampler;

/// THREE LIGHT SHADOW UNIFORMS AND VARYINGS
//we dont do shdows if there are 8 textures units, so wrapping this will enable us to stay
//under total number of texture units and the total number of uniforms that is also limited on ios13
#ifndef NUM_TEXTURE_UNITS_8
uniform float keyShadowDarkness;
varying vec4 v_PositionFromKeyLight;
varying float v_KeyLightDepthMetric;
uniform sampler2D _KeyLightShadowSampler;
uniform float backShadowDarkness;
varying vec4 v_PositionFromBackLight;
varying float v_BackLightDepthMetric;
uniform sampler2D _BackLightShadowSampler;
uniform float fillShadowDarkness;
varying vec4 v_PositionFromFillLight;
varying float v_FillLightDepthMetric;
uniform sampler2D _FillLightShadowSampler;
#endif

// Used to select the tone mapping function.
// Here is a list of the function with their index. Note These values
// match the ToneMapType enum in the applicaiton.
//
//    none                                              = 0
//    filmicToneMapping                                 = 1,
//    linearToneMapping                                 = 2,
//    simpleReinhardToneMapping                         = 3,
//    lumaBasedReinhardToneMapping                      = 4,
//    whitePreservingLumaBasedReinhardToneMapping       = 5,
//    RomBinDaHouseToneMapping                          = 6,
//    Uncharted2ToneMapping                             = 7,
//    BHF_Special                                       = 8
uniform int _ToneMappingFunction;


// Uniforms properties
uniform vec2 indirectMaskValues;

// Albedo/Diffuse
uniform vec4 _Color;
uniform float _Cutoff;
uniform sampler2D _MainTex;
uniform float hasAlbedo;

//Albedo/Roughness Masking
uniform vec4 _MaskColor;
uniform float _MaskRoughness;
uniform float _UseMask;

// Ambient Occlusion
uniform sampler2D _OcclusionMap;
uniform float _OcclusionStrength;
uniform float _AOSecondTileset;
uniform float hasOcclusion;

// Metallic Glossiness
uniform float _Metallic;
uniform float _Glossiness;
uniform sampler2D _MetallicGlossMap;
uniform float _MSecondTileset;
uniform float hasMetallicGloss;

// Base Normals
#ifdef NORMAL_MAP
uniform sampler2D _BumpMap;
#endif /// NORMAL_MAP ///
uniform float _BumpScale;

// Emission
uniform sampler2D _EmissionMap;
uniform vec3 _EmissionColor;
uniform float hasEmission;

// Detail Albedo
uniform sampler2D _DetailAlbedoMap;
uniform float hasDetailAlbedo;

// Detail Normal
#ifdef DETAIL_NORMAL
uniform sampler2D _DetailNormalMap;
#endif /// DETAIL_NORMAL ///

// Detail Utility
uniform float _DetailAlpha;
uniform float _DetailBumpScale;
uniform float _DetailMultiply;

// Environment Reflection Maps
uniform samplerCube _EnvironmentMapLOD0;
uniform samplerCube _EnvironmentMapLOD1;
uniform samplerCube _EnvironmentMapLOD2;
uniform samplerCube _EnvironmentMapLOD3;
uniform vec3 _ProbeHalfSize;
uniform vec3 _BoxCenter;
uniform float _ReflectionProbeStrength;

/// Box Projection Constants
const vec3 BOX_CENTER = vec3(0.0, 1.4, 0.0);

/// Shader uniform constants
const vec4 DIELECTRICSPEC = vec4(0.04, 0.04, 0.04, 1.0 - 0.04);
const vec4 COLORSPACEDOUBLE = vec4(4.59479380, 4.59479380, 4.59479380, 2.0);

/// Ambient Colors
uniform vec3 AMBIENTCOLORHIGH;
uniform vec3 AMBIENTCOLORMID;
uniform vec3 AMBIENTCOLORLOW;

/// Light properties
#ifdef DIRECTIONAL_LIGHTING
/// Directional Light properties ///
uniform mat4 LIGHT_DIRECTIONAL_PARAMS;
/// LIGHT COLOR
/// LIGHT DIRECTION ///
/// LIGHT INTENSITY, SPECULAR INTENSITY ///
#endif /// DIRECTIONAL_LIGHTING ///

/// Spot Light properties ///
#ifdef SPOT_LIGHTING
uniform mat4 LIGHT_SPOT_PARAMS;
/// LIGHT COLOR, SPEC ATTENUTATION
/// LIGHT DIRECTION, LIGHT height ///
/// LIGHT X SPACING, 1/2 LIGHT X SPACING, LIGHT Y SPACING, 1/2 LIGHT Y SPACING ///
/// LIGHT RANGE, 1/LIGHT RANGE, PHIDOT, 1/max(.01, THETADOT - PHIDOT) ///
#endif /// SPOT_LIGHTING ///


#ifdef THREEPOINTLIGHTING
uniform mat4 KEY_LIGHT_SPOT_PARAMS;
uniform mat4 FILL_LIGHT_SPOT_PARAMS;
uniform mat4 BACK_LIGHT_SPOT_PARAMS;
/// LIGHT COLOR, SPEC ATTENUTATION
/// LIGHT DIRECTION, LIGHT height ///
/// LIGHT X SPACING, 1/2 LIGHT X SPACING, LIGHT Y SPACING, 1/2 LIGHT Y SPACING ///
/// LIGHT RANGE, 1/LIGHT RANGE, PHIDOT, 1/max(.01, THETADOT - PHIDOT) ///
#endif /// THREEPOINTLIGHTING ///


#ifdef WALL_SPOT_LIGHTING
uniform vec4 WALL_SPOT_LIGHT_COLOR_PARAMS; /// WALL LIGHT COLOR, SPEC ATTENUATION ///

/// WALL NORMAL X, WALL NORMAL Z, WALL SEGMENTLENGTH, 1/WALL SEGMENTLENGTH ///
/// WALL END POINT 1 XZ, WALL LIGHT SEGMENT COUNT - 1, WALL LIGHT HEIGHT ///
/// WALL END POINT 2 XZ, 1/WALL LIGHT SEGMENT COUNT, WALL OFFSET ///
/// WALL LIGHT RANGE, 1/WALL LIGHT RANGE, PHIDOT, 1/max(0.01, THETADOT - PHIDOT) ///
uniform mat4 WALL_SPOT_LIGHT_PARAMS;
#endif /// WALL_SPOT_LIGHTING ///

#ifdef WALL_POINT_LIGHTING
uniform vec3 WALL_POINT_LIGHT_COLOR_PARAMS; /// WALL LIGHT COLOR ///

/// WALL NORMAL X, WALL NORMAL Z, WALL SEGMENTLENGTH, 1/WALL SEGMENTLENGTH ///
/// WALL END POINT 1 XZ, WALL LIGHT SEGMENT COUNT - 1, WALL LIGHT HEIGHT ///
/// WALL END POINT 2 XZ, 1/WALL LIGHT SEGMENT COUNT, WALL OFFSET ///
/// WALL LIGHT RANGE, 1/WALL LIGHT RANGE ///
uniform mat4 WALL_POINT_LIGHT_PARAMS;

#endif ///

vec4 GammaToLinearSpace (vec4 sRGB)
{
    return vec4(sRGB.rgb * (sRGB.rgb * (sRGB.rgb * 0.305306011 + 0.682171111) + 0.012522878), sRGB.a);
}

vec3 LinearToGammaSpace (vec3 linRGB)
{
    linRGB = max(linRGB, vec3(0.0, 0.0, 0.0));
    // An almost-perfect approximation from http://chilliant.blogspot.com.au/2012/08/srgb-approximations-for-hlsl.html?m=1
    return max(1.055 * pow(linRGB, vec3(0.416666667, 0.416666667, 0.416666667)) - 0.055, 0.0);
}

vec3 LerpWhiteTo(vec3 b, float t)
{
    return vec3(1.0, 1.0, 1.0) + t * (b - vec3(1.0, 1.0, 1.0));
}

vec3 UnpackScaleNormalRGorAG(vec4 packedNormal, float bumpScale)
{
    packedNormal.x *= packedNormal.w;
    vec3 normal;
    normal.xy = packedNormal.xy * 2.0 - 1.0;
    normal.xy *= bumpScale;
    normal.z = sqrt(1.0 - clamp(dot(normal.xy, normal.xy), 0.0, 1.0));
    return normal;
}

vec3 Vec3TsToWsNormalized( vec3 vVectorTs, vec3 vNormalWs, vec3 vTangentUWs, vec3 vTangentVWs )
{
	return (normalize(
		//start by mul the horizontal channel of the normal sample with the tangent of the vertex.
		vVectorTs.x * vTangentUWs.xyz
		//mul the vertical channel of the normal sample with the binormal of the vertex.
		+vVectorTs.y * vTangentVWs.xyz
		//mul the depth channel of the normal sample with the normal of the vertex.
		+vVectorTs.z * vNormalWs.xyz
	));
}

float OneMinusReflectivityFromMetallic(float metallic)
{
	return DIELECTRICSPEC.a - metallic * DIELECTRICSPEC.a;
}

vec3 SpecularFromMetallic (vec3 albedo, float metallic)
{
	return mix (DIELECTRICSPEC.rgb, albedo, metallic);
}

vec3 BRDF1_Unity_PBS (vec3 specColor_directionalcolor, float roughness, float a2, vec3 normal, vec3 _halfv, float ldoth, float ndotl)
{
	float nh = dot(normal, _halfv);
    float d = nh * nh * (a2 - 1.0) + 1.00001;
	float specularTerm = a2 / (max(0.1, ldoth*ldoth) * (roughness + 0.5) * (d * d) * 4.0);
	specularTerm = specularTerm - 1e-4;
	specularTerm = clamp(specularTerm, 0.0, 100.0); //Prevent FP16 overflow on mobiles
	return specularTerm * specColor_directionalcolor * ndotl;
}

vec3 FresnelLerpFast (vec3 F0, vec3 F90, float cosA)
{
    float oneMinus = 1.0 - cosA;
	float t = oneMinus * oneMinus * oneMinus * oneMinus;
	return mix (F0, F90, t);
}

vec3 DecodeHDR (vec4 data, vec4 decodeInstructions)
{
	float alpha = decodeInstructions.w * (data.a - 1.0) + 1.0;

	return (decodeInstructions.x * pow(alpha, decodeInstructions.y)) * data.rgb;
}

float Enforce0or1(float value)
{
    // For some reason the Mac M1's clamp function is returning values slightly beyond the clamp value or slightly near it.
    // But the glossy cubemap switch matrix requires absolutely exact values of 0.0 or 1.0 to function correctly, so here we 
    // enforce the values to be exactly 0.0 or 1.0.
    float myOutput = value;
    if (value >= 0.5)
        myOutput = 1.0;
    else
        myOutput = 0.0;
    return myOutput;
}

vec3 Unity_GlossyEnvironment (vec3 reflUVW, float roughness, vec4 decodeInstructions)
{
	roughness = roughness * (1.7 - 0.7 * roughness);
	float mip = clamp(roughness * 3.0, 0.0, 3.0);
    float mipFloor = floor(mip);
    float mipCeil = ceil(mip);

    #ifdef NUM_TEXTURE_UNITS_8
    vec4 reflColor = textureCube(_EnvironmentMapLOD3, reflUVW, 0.0);
    mat4 lods = mat4(reflColor,
                    reflColor,
                    reflColor,
                    reflColor);
    #else
    mat4 lods = mat4(textureCube(_EnvironmentMapLOD0, reflUVW, 0.0),
                    textureCube(_EnvironmentMapLOD1, reflUVW, 0.0),
                    textureCube(_EnvironmentMapLOD2, reflUVW, 0.0),
                    textureCube(_EnvironmentMapLOD3, reflUVW, 0.0));
    #endif
    
    vec4 floors = clamp(vec4(1.0, mipFloor, mipFloor - 1.0, mipFloor - 2.0), vec4(0.0, 0.0, 0.0, 0.0), vec4(1.0, 1.0, 1.0, 1.0));
    floors.x = Enforce0or1(floors.x);
    floors.y = Enforce0or1(floors.y);
    floors.z = Enforce0or1(floors.z);
    floors.w = Enforce0or1(floors.w);
    floors.x -= floors.y;
    floors.y -= floors.z;
    floors.z -= floors.w;
    vec4 rgbFloor = lods * floors;

    vec4 ceils = clamp(vec4(1.0, mipCeil, mipCeil - 1.0, mipCeil - 2.0), vec4(0.0, 0.0, 0.0, 0.0), vec4(1.0, 1.0, 1.0, 1.0));
    ceils.x = Enforce0or1(ceils.x);
    ceils.y = Enforce0or1(ceils.y);
    ceils.z = Enforce0or1(ceils.z);
    ceils.w = Enforce0or1(ceils.w);
    ceils.x -= ceils.y;
    ceils.y -= ceils.z;
    ceils.z -= ceils.w;
    vec4 rgbCeil = lods * ceils;

    vec4 rgbm = mix(rgbFloor, rgbCeil, mip - mipFloor);
	return GammaToLinearSpace(rgbm).rgb;
}

vec3 BoxProjectedCubemapDirection(vec3 worldRefl, vec3 worldPos)
{
	// Find the ray intersection with box plane
	vec3 invOrigVec = vec3(1.0,1.0,1.0) / worldRefl;
	vec3 intersecAtMaxPlane = (_BoxCenter + _ProbeHalfSize - worldPos) * invOrigVec;
	vec3 intersecAtMinPlane = (_BoxCenter - _ProbeHalfSize - worldPos) * invOrigVec;
	// Get the largest intersection values (we are not intersted in negative values)
	vec3 largestIntersec = max(intersecAtMaxPlane, intersecAtMinPlane);
	// Get the closest of all solutions
	float distance = min(min(largestIntersec.x, largestIntersec.y), largestIntersec.z);
	// Get the intersection position
	vec3 intersectPositionWS = worldPos + worldRefl * distance;
	// Get corrected vector
	return normalize(intersectPositionWS - _BoxCenter);
}

vec3 CalculateIndirect(float roughness, float realRoughness, vec3 normal, float oneMinusReflectivity, vec3 specColor, vec3 worldPos, float ndotv, vec3 viewDir, float reflectionPower)
{
    if(reflectionPower == 0.0)
    {
        return vec3(0.0,0.0,0.0);
    }

	float grazingTerm = clamp((1.0 - roughness) + (1.0 - oneMinusReflectivity), 0.0, 1.0);

	float surfaceReduction = 1.0 - realRoughness * roughness * (0.6 - 0.08 * roughness);

	vec3 worldRefl = reflect(-viewDir, normal);
    worldRefl = BoxProjectedCubemapDirection(worldRefl, worldPos);
    worldRefl.y *= -1.0;

	vec4 decode_instructions = vec4(1.0, 1.0, 0.0, 0.0); // x is ReflectionIntensity

    vec3 reflData = Unity_GlossyEnvironment(worldRefl, roughness, decode_instructions);

    return clamp(surfaceReduction * reflData * FresnelLerpFast(specColor, vec3(grazingTerm, grazingTerm, grazingTerm), ndotv) * reflectionPower, 0.0, 1.0);
}

vec3 PreMultiplyAlpha (vec3 diffColor, float alpha, float oneMinusReflectivity, out float outModifiedAlpha)
{
	#ifdef ALPHA_PREMULTIPLY
        diffColor = LinearToGammaSpace(diffColor) * alpha;
        diffColor = GammaToLinearSpace(vec4(diffColor, alpha)).rgb;
        outModifiedAlpha = 1.0 - oneMinusReflectivity + alpha * oneMinusReflectivity;
    #else
    	outModifiedAlpha = alpha;
    #endif /// ALPHA_PREMULTIPLY ///

 	return diffColor;
}

float unpack(vec4 color)
{
    const vec4 bit_shift = vec4(1.0 / (255.0 * 255.0 * 255.0), 1.0 / (255.0 * 255.0), 1.0 / 255.0, 1.0);
    return dot(color, bit_shift);
}

float computeDirectionalShadow(vec4 v_PositionFromLight, sampler2D shadowSampler, float darkness)
{
    vec3 depth = v_PositionFromLight.xyz / v_PositionFromLight.w;
    depth = 0.5 * depth + vec3(0.5);
    vec2 uv = depth.xy;

    if (uv.x < 0. || uv.x > 1.0 || uv.y < 0. || uv.y > 1.0)
    {
        return 1.0;
    }

    #ifndef SHADOWFULLFLOAT
        float shadow = unpack(texture2D(shadowSampler, uv));
    #else
        float shadow = texture2D(shadowSampler, uv).x;
    #endif

    if (depth.z > shadow)
    {
        return 1.0 - darkness;
    }

    return 1.0;
}

float computeSpotShadow(vec4 v_PositionFromLight, float depthMetric, sampler2D shadowSampler, float darkness)
{
    vec3 depth = v_PositionFromLight.xyz / v_PositionFromLight.w;
    //move from -1->1 range to 0->1 range
    depth = 0.5 * depth + vec3(0.5);
    vec2 uv = depth.xy;

    if (uv.x < 0. || uv.x > 1.0 || uv.y < 0. || uv.y > 1.0)
    {
        return 1.0;
    }

    float shadowPixelDepth = clamp(depthMetric, 0., 1.0);

    #ifndef SHADOWFULLFLOAT
        float shadow = unpack(texture2D(shadowSampler, uv));
    #else
        float shadow = texture2D(shadowSampler, uv).x;
    #endif

    if (shadowPixelDepth > shadow)
    {
        return 1.0 - darkness;
    }

    return 1.0;
}
float gamma = 2.2;
vec3 linearToneMapping(vec3 color)
{
	float exposure = 1.0;
	color = clamp(exposure * color, 0.0, 1.0);
	color = pow(color, vec3(1.0 / gamma));
	return color;
}
vec3 simpleReinhardToneMapping(vec3 color)
{
	float exposure = 1.5;
	color *= exposure/(1.0 + color / exposure);
	color = pow(color, vec3(1.0 / gamma));
	return color;
}

vec3 lumaBasedReinhardToneMapping(vec3 color)
{
	float luma = dot(color, vec3(0.2126, 0.7152, 0.0722));
	float toneMappedLuma = luma / (1.0 + luma);
	color *= toneMappedLuma / luma;
	color = pow(color, vec3(1.0 / gamma));
	return color;
}

vec3 whitePreservingLumaBasedReinhardToneMapping(vec3 color)
{
	float white = 2.0;
	float luma = dot(color, vec3(0.2126, 0.7152, 0.0722));
	float toneMappedLuma = luma * (1.0 + luma / (white*white)) / (1.0 + luma);
	color *= toneMappedLuma / luma;
	color = pow(color, vec3(1.0 / gamma));
	return color;
}

vec3 RomBinDaHouseToneMapping(vec3 color)
{
    color = exp( -1.0 / ( 2.72*color + 0.15 ) );
	color = pow(color, vec3(1.0 / gamma));
	return color;
}

vec3 filmicToneMapping(vec3 color)
{
	color = max(vec3(0.0), color - vec3(0.004));
	color = (color * (6.2 * color + 0.5)) / (color * (6.2 * color + 1.7) + 0.06);
	return color;
}

vec3 Uncharted2ToneMapping(vec3 color)
{
	float A = 0.15;
	float B = 0.50;
	float C = 0.10;
	float D = 0.20;
	float E = 0.02;
	float F = 0.30;
	float W = 11.2;
	float exposure = 2.0;
	color *= exposure;
	color = ((color * (A * color + C * B) + D * E) / (color * (A * color + B) + D * F)) - E / F;
	float white = ((W * (A * W + C * B) + D * E) / (W * (A * W + B) + D * F)) - E / F;
	color /= white;
	color = pow(color, vec3(1.0 / gamma));
	return color;
}
void main() {
   
    vec2 uv = v_uv.xy;
    vec2 detailUV = v_uv.zw;
    /// Setup Albedo and Alpha
    vec4 albedo = GammaToLinearSpace(_Color);
    //store the OS albedo before we do anything to it in case of masking
    vec4 os_albedo = texture2D(_MainTex, uv);
    albedo = mix(albedo, GammaToLinearSpace(os_albedo) * albedo, hasAlbedo); // WebGL 1.0
    float outputAlpha = albedo.a;

    /// Initialize emission ///
    vec3 emission = GammaToLinearSpace(vec4(_EmissionColor,1.0)).rgb;
    #ifdef EMISSION_LUM
    //Set emission not using the color tint
    //This way, if we have this toggle on, it will always render regardless of color
        emission = mix(emission, GammaToLinearSpace(texture2D(_EmissionMap, uv)).rgb, hasEmission); // WebGL 1.0
    #else
    //If we turn the toggle off, it should render like a normal emission map with a color tint
    //Set emission using the color tint
        emission = mix(emission, GammaToLinearSpace(texture2D(_EmissionMap, uv)).rgb * emission, hasEmission); // WebGL 1.0
    #endif

#ifdef ALPHA_TEST
    if(outputAlpha < _Cutoff)
        discard;
#endif /// ALPHA_TEST ///

    /// Setup Detail Albedo
    vec4 detailSample = mix(vec4(0.5,0.5,0.5,1.0), GammaToLinearSpace(texture2D(_DetailAlbedoMap, detailUV)), hasDetailAlbedo); // WebGL 1.0
    vec3 detailColor = albedo.rgb * mix(LerpWhiteTo(detailSample.rgb * COLORSPACEDOUBLE.rgb, detailSample.a), detailSample.rgb, _DetailMultiply);
    albedo.rgb = mix(albedo.rgb, detailColor, hasDetailAlbedo);    
	outputAlpha = mix(outputAlpha, detailSample.a, _DetailAlpha);

    /// Setup Normals
    vec3 normalWS = v_normal;
    float _bumpScale;
#ifdef NORMAL_MAP
    _bumpScale = mix(_BumpScale, 1.0, _DetailBumpScale);
    vec3 normalTS = UnpackScaleNormalRGorAG(texture2D(_BumpMap, uv), _bumpScale); // WebGL 1.0
#else
    vec3 normalTS = vec3(0.0,0.0,1.0);
#endif /// NORMAL_MAP ///

#ifdef DETAIL_NORMAL
    _bumpScale = mix(1.0, _BumpScale, _DetailBumpScale);
    vec3 detailNormal = UnpackScaleNormalRGorAG(texture2D(_DetailNormalMap, detailUV), _BumpScale); // WebGL 1.0
    normalTS = normalize(vec3(normalTS.xy + detailNormal.xy, normalTS.z * detailNormal.z));
#endif /// DETAIL_NORMAL ///
    normalWS = Vec3TsToWsNormalized(normalTS, normalWS, v_tangentuws, v_tangentvws);
    /// Setup Occlusion /// 
    vec2 oUV = mix(uv, detailUV, _AOSecondTileset);
    vec4 aoSample = texture2D(_OcclusionMap, oUV); 
    float occlusion = mix(1.0, mix(1.0, mix(1.0, aoSample.r, _OcclusionStrength), clamp(indirectMaskValues.x, 0.0, 1.0)), hasOcclusion); // WebGL 1.0
    
	/// Initialize Diffuse and specular terms ///
	vec3 diffuseTerm = vec3(0.0, 0.0, 0.0);
	vec3 specTerm = vec3(0.0, 0.0, 0.0);
	vec3 viewDir = normalize(v_view - v_position);
	float metallic = 0.0;
    float roughness = 0.0;
    float realRoughness = 0.0;
    vec3 specColor = vec3(0.0,0.0,0.0);
    float oneMinusReflectivity = 0.0;
    /// PERFORM LIGHTING IF APPROPRIATE MODE ///
    #if defined(DIRECTIONAL_LIGHTING) || defined(SPOT_LIGHTING) || defined(WALL_SPOT_LIGHTING) || defined(WALL_POINT_LIGHTING) || defined(THREEPOINTLIGHTING)

    /// Setup Metallic Roughness ///
    vec2 mUV = mix(uv, detailUV, _MSecondTileset);
    vec4 metallicTexel = texture2D(_MetallicGlossMap, mUV); // WebGL 1.0
    metallic = mix(_Metallic, metallicTexel.r * _Metallic, hasMetallicGloss);
    
    /// Setup Roughness ///
    roughness = max(mix(1.0 - _Glossiness, 1.0 - (metallicTexel.a * _Glossiness), hasMetallicGloss), 0.01);
    
    /// Setup Optional Mask ///
    if(_UseMask == 1.0){
        roughness = mix(roughness, _MaskRoughness, aoSample.g);
        albedo.rgb = mix(albedo.rgb, GammaToLinearSpace(_MaskColor).rgb, aoSample.g);
    }
 
    realRoughness = roughness * roughness;
    realRoughness *= realRoughness;
    specColor = SpecularFromMetallic(albedo.rgb, metallic);

    /// Setup OneMinusReflectivity ///
    oneMinusReflectivity = OneMinusReflectivityFromMetallic(metallic);
    albedo *= oneMinusReflectivity;
    
    /// premultiply alpha before lighting calculations ///
    albedo.rgb = PreMultiplyAlpha(albedo.rgb, outputAlpha, oneMinusReflectivity, outputAlpha);

#if defined(NUM_TEXTURE_UNITS_8)
    #if defined(ISOUTSIDE)
        float directionalShadowTerm = 1.0;
    #else
        float directionalShadowTerm = 0.0;
    #endif
    float spotShadowTerm = 1.0;
    float keySpotShadowTerm = 1.0;
    float backSpotShadowTerm = 1.0;
    float fillSpotShadowTerm = 1.0;
#else
    #if defined(DIRECTIONAL_LIGHTING)
        float directionalShadowTerm = computeDirectionalShadow(v_PositionFromDirLight, _DirectionalShadowSampler, shadowDarkness);
    #endif
    #if defined(SPOT_LIGHTING) || defined(WALL_SPOT_LIGHTING)
        float spotShadowTerm = computeSpotShadow(v_PositionFromSpotLight, v_DepthMetric, _SpotLightShadowSampler, shadowDarkness);
    #elif defined(THREEPOINTLIGHTING)
        float keySpotShadowTerm = computeSpotShadow(v_PositionFromKeyLight, v_KeyLightDepthMetric, _KeyLightShadowSampler, keyShadowDarkness);
        float backSpotShadowTerm = computeSpotShadow(v_PositionFromBackLight, v_BackLightDepthMetric, _BackLightShadowSampler, backShadowDarkness);
        float fillSpotShadowTerm = computeSpotShadow(v_PositionFromFillLight, v_FillLightDepthMetric, _FillLightShadowSampler, fillShadowDarkness);
    #endif
#endif
    
    /// determine lighting effects to use ///
    #ifdef DIRECTIONAL_LIGHTING
    /// Setup Lighting specific vectors
    vec3 lightDir = -normalize(LIGHT_DIRECTIONAL_PARAMS[1].xyz);
    float ndotld = max(0.0, dot(normalWS, lightDir));
    vec3 _halfvd = normalize(lightDir + viewDir);
    float ldothd = max(0.0, dot(lightDir, _halfvd)); 
    /// Calculate Diffuse and specular terms
    #ifdef EMISSION_LUM
    directionalShadowTerm = occlusion;
    #endif
    diffuseTerm += ndotld * LIGHT_DIRECTIONAL_PARAMS[0].xyz * LIGHT_DIRECTIONAL_PARAMS[2].x * directionalShadowTerm;
    specTerm += BRDF1_Unity_PBS(specColor * LIGHT_DIRECTIONAL_PARAMS[0].xyz, roughness, realRoughness, normalWS, _halfvd, ldothd, ndotld) * LIGHT_DIRECTIONAL_PARAMS[2].y * directionalShadowTerm;
    #endif /// DIRECTIONAL_LIGHTING ///

    #if defined(THREEPOINTLIGHTING)

    #ifdef EMISSION_LUM
    keySpotShadowTerm = occlusion;
    backSpotShadowTerm = occlusion;
    fillSpotShadowTerm = occlusion;
    #endif

    // Key Light
    vec3 spotspeccolor = specColor * KEY_LIGHT_SPOT_PARAMS[0].xyz;
    vec3 l2fdir = (KEY_LIGHT_SPOT_PARAMS[2].xyz - v_position);
    float l2f = length(l2fdir);
    l2fdir = normalize(l2fdir);
    float ndotl = max(0.0, dot(normalWS, l2fdir));
    float ldotd = dot(KEY_LIGHT_SPOT_PARAMS[1].xyz, -l2fdir);
    float attenuation = (1.0 - min(l2f * KEY_LIGHT_SPOT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - KEY_LIGHT_SPOT_PARAMS[3].z) * KEY_LIGHT_SPOT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    float spottest = clamp(sign(KEY_LIGHT_SPOT_PARAMS[3].x - l2f) + sign(ldotd - KEY_LIGHT_SPOT_PARAMS[3].z), 0.0, 1.0);
    vec3 keyDiffuse = mix(vec3(0.0, 0.0, 0.0), ndotl * KEY_LIGHT_SPOT_PARAMS[0].xyz * attenuation, spottest);
    vec3 keyDiffuseWithShadow = keyDiffuse * keySpotShadowTerm;
    diffuseTerm += keyDiffuseWithShadow;

    vec3 _halfv = normalize(l2fdir + viewDir);
    float ldoth = max(0.0, dot(l2fdir, _halfv));
    vec3 keySpecTerm = mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * KEY_LIGHT_SPOT_PARAMS[0].w, spottest);
    vec3 keySpecTermWithShadow = keySpecTerm * keySpotShadowTerm;
    specTerm += keySpecTermWithShadow;

    // Fill Light
    spotspeccolor = specColor * FILL_LIGHT_SPOT_PARAMS[0].xyz;
    l2fdir = (FILL_LIGHT_SPOT_PARAMS[2].xyz - v_position);
    l2f = length(l2fdir);
    l2fdir = normalize(l2fdir);
    ndotl = max(0.0, dot(normalWS, l2fdir));
    ldotd = dot(FILL_LIGHT_SPOT_PARAMS[1].xyz, -l2fdir);
    attenuation = (1.0 - min(l2f * FILL_LIGHT_SPOT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - FILL_LIGHT_SPOT_PARAMS[3].z) * FILL_LIGHT_SPOT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    spottest = clamp(sign(FILL_LIGHT_SPOT_PARAMS[3].x - l2f) + sign(ldotd - FILL_LIGHT_SPOT_PARAMS[3].z), 0.0, 1.0);
    vec3 fillDiffuse = mix(vec3(0.0, 0.0, 0.0), ndotl * FILL_LIGHT_SPOT_PARAMS[0].xyz * attenuation, spottest);
    vec3 fillDiffuseWithShadow = fillDiffuse * fillSpotShadowTerm;
    diffuseTerm += fillDiffuseWithShadow;

    _halfv = normalize(l2fdir + viewDir);
    ldoth = max(0.0, dot(l2fdir, _halfv));
    vec3 fillSpecTerm = mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * FILL_LIGHT_SPOT_PARAMS[0].w, spottest);
    vec3 fillSpecTermWithShadow = fillSpecTerm * fillSpotShadowTerm;
    specTerm += fillSpecTermWithShadow;

    // Back Light
    spotspeccolor = specColor * BACK_LIGHT_SPOT_PARAMS[0].xyz;
    l2fdir = (BACK_LIGHT_SPOT_PARAMS[2].xyz - v_position);
    l2f = length(l2fdir);
    l2fdir = normalize(l2fdir);
    ndotl = max(0.0, dot(normalWS, l2fdir));
    ldotd = dot(BACK_LIGHT_SPOT_PARAMS[1].xyz, -l2fdir);
    attenuation = (1.0 - min(l2f * BACK_LIGHT_SPOT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - BACK_LIGHT_SPOT_PARAMS[3].z) * BACK_LIGHT_SPOT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    spottest = clamp(sign(BACK_LIGHT_SPOT_PARAMS[3].x - l2f) + sign(ldotd - BACK_LIGHT_SPOT_PARAMS[3].z), 0.0, 1.0);
    vec3 backDiffuse = mix(vec3(0.0, 0.0, 0.0), ndotl * BACK_LIGHT_SPOT_PARAMS[0].xyz * attenuation, spottest);
    vec3 backDiffuseWithShadow = backDiffuse * backSpotShadowTerm;
    diffuseTerm += backDiffuseWithShadow;

    _halfv = normalize(l2fdir + viewDir);
    ldoth = max(0.0, dot(l2fdir, _halfv));
    vec3 backSpecTerm = mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * BACK_LIGHT_SPOT_PARAMS[0].w, spottest);
    vec3 backSpecTermWithShadow = backSpecTerm * backSpotShadowTerm;
    specTerm += backSpecTermWithShadow;
    #endif /// THREEPOINTLIGHTING ///


    #ifdef SPOT_LIGHTING

    /// Always 4 Lights ///
    /// first determine which "zone" the fragment is in, using the fragment's X and Z position ///
    vec2 center = (floor(v_position.xz * LIGHT_SPOT_PARAMS[2].yw) + 0.5) * LIGHT_SPOT_PARAMS[2].xz;
    vec2 halfcenter = LIGHT_SPOT_PARAMS[2].xz*0.5;

    /// spec color is the same for every light ///
    vec3 spotspeccolor = specColor * LIGHT_SPOT_PARAMS[0].xyz;

    /// direction of ray going from the fragment to the light position
    vec3 l2fdir = (vec3(center.x - halfcenter.x, LIGHT_SPOT_PARAMS[1].w, center.y - halfcenter.y) - v_position);
    /// distance between fragment and light ///
    float l2f = length(l2fdir);
    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);
    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    float ndotl = max(0.0, dot(normalWS, l2fdir));
    /// dot of the light direction and the direction from the light to the fragment ///
    float ldotd = dot(LIGHT_SPOT_PARAMS[1].xyz, -l2fdir);
    /// attenuate the light strength based on the distance to the light and the light range ///
    float attenuation = (1.0 - min(l2f * LIGHT_SPOT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - LIGHT_SPOT_PARAMS[3].z) * LIGHT_SPOT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    float spottest = clamp(sign(LIGHT_SPOT_PARAMS[3].x - l2f) + sign(ldotd - LIGHT_SPOT_PARAMS[3].z), 0.0, 1.0);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * LIGHT_SPOT_PARAMS[0].xyz * attenuation, spottest);

    vec3 _halfv = normalize(l2fdir + viewDir);
    float ldoth = max(0.0, dot(l2fdir, _halfv));
    specTerm += mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * LIGHT_SPOT_PARAMS[0].w, spottest);


    l2fdir = (vec3(center.x - halfcenter.x, LIGHT_SPOT_PARAMS[1].w, center.y + halfcenter.y) - v_position);
    l2f = length(l2fdir);
    l2fdir = normalize(l2fdir);
    ndotl = max(0.0, dot(normalWS, l2fdir));
    ldotd = dot(LIGHT_SPOT_PARAMS[1].xyz, -l2fdir);
    attenuation = (1.0 - min(l2f * LIGHT_SPOT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - LIGHT_SPOT_PARAMS[3].z) * LIGHT_SPOT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    spottest = clamp(sign(LIGHT_SPOT_PARAMS[3].x - l2f) + sign(ldotd - LIGHT_SPOT_PARAMS[3].z), 0.0, 1.0);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * LIGHT_SPOT_PARAMS[0].xyz * attenuation, spottest);

    _halfv = normalize(l2fdir + viewDir);
    ldoth = max(0.0, dot(l2fdir, _halfv));
    specTerm += mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * LIGHT_SPOT_PARAMS[0].w, spottest);
    

    l2fdir = (vec3(center.x + halfcenter.x, LIGHT_SPOT_PARAMS[1].w, center.y - halfcenter.y) - v_position);
    l2f = length(l2fdir);
    l2fdir = normalize(l2fdir);
    ndotl = max(0.0, dot(normalWS, l2fdir));
    ldotd = dot(LIGHT_SPOT_PARAMS[1].xyz, -l2fdir);
    attenuation = (1.0 - min(l2f * LIGHT_SPOT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - LIGHT_SPOT_PARAMS[3].z) * LIGHT_SPOT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    spottest = clamp(sign(LIGHT_SPOT_PARAMS[3].x - l2f) + sign(ldotd - LIGHT_SPOT_PARAMS[3].z), 0.0, 1.0);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * LIGHT_SPOT_PARAMS[0].xyz * attenuation, spottest);

    _halfv = normalize(l2fdir + viewDir);
    ldoth = max(0.0, dot(l2fdir, _halfv));
    specTerm += mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * LIGHT_SPOT_PARAMS[0].w, spottest);


    l2fdir = (vec3(center.x + halfcenter.x, LIGHT_SPOT_PARAMS[1].w, center.y + halfcenter.y) - v_position);
    l2f = length(l2fdir);
    l2fdir = normalize(l2fdir);
    ndotl = max(0.0, dot(normalWS, l2fdir));
    ldotd = dot(LIGHT_SPOT_PARAMS[1].xyz, -l2fdir);
    attenuation = (1.0 - min(l2f * LIGHT_SPOT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - LIGHT_SPOT_PARAMS[3].z) * LIGHT_SPOT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    spottest = clamp(sign(LIGHT_SPOT_PARAMS[3].x - l2f) + sign(ldotd - LIGHT_SPOT_PARAMS[3].z), 0.0, 1.0);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * LIGHT_SPOT_PARAMS[0].xyz * attenuation, spottest);
    #ifdef EMISSION_LUM
    spotShadowTerm = occlusion;
    #endif
    diffuseTerm *= spotShadowTerm;

    _halfv = normalize(l2fdir + viewDir);
    ldoth = max(0.0, dot(l2fdir, _halfv));
    specTerm += mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * LIGHT_SPOT_PARAMS[0].w, spottest);
    specTerm *= spotShadowTerm;

    #elif defined(WALL_SPOT_LIGHTING)
    /// spec color is the same for every light ///
    vec3 spotspeccolor = specColor * WALL_SPOT_LIGHT_COLOR_PARAMS.xyz;

    /// determine the number of lights, which is the wall width / light spacing ///
    ///float lightCount = WALL_SPOT_LIGHT_PARAMS[0].y;
    /// determine which section the fragment is in ...
    vec3 wallStartPoint = vec3(WALL_SPOT_LIGHT_PARAMS[1].x, 0.0, WALL_SPOT_LIGHT_PARAMS[1].y);
    vec3 wallDirection = normalize(vec3(WALL_SPOT_LIGHT_PARAMS[2].x, 0.0, WALL_SPOT_LIGHT_PARAMS[2].y) - wallStartPoint);
    
    /// need to determine the point along the wall corresponding to the vertex ///
    /// find distance from wallStartPoint to vertex (at the same height as wallStartPoint ///
    vec3 vertexWallDirection = (vec3(v_position.x, 0.0, v_position.z) - wallStartPoint);
    ///dot product of wallDirection and vertexWallDirection * length of vertexWallDirection ///
    float vertexWallDistance = dot(wallDirection, normalize(vertexWallDirection))*length(vertexWallDirection);
    float segment = floor(vertexWallDistance * WALL_SPOT_LIGHT_PARAMS[0].w);
    
    /// Always 4 Lights ///
    /// determine the position of the light toward point 1 of the wall ///
    /// offset from the wall along the wall normal ///
    vec3 lightPos = wallStartPoint + wallDirection * WALL_SPOT_LIGHT_PARAMS[0].z * segment + vec3(WALL_SPOT_LIGHT_PARAMS[0].x, 0.0, WALL_SPOT_LIGHT_PARAMS[0].y)*WALL_SPOT_LIGHT_PARAMS[2].w;
    lightPos.y = WALL_SPOT_LIGHT_PARAMS[1].w;
    /// perform the light calculations ///
    /// direction of ray going from the fragment to the light position
    vec3 l2fdir = (lightPos - v_position);

    /// distance between fragment and light ///
    float l2f = length(l2fdir);

    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);

    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    float ndotl = max(0.0, dot(normalWS, l2fdir));

    /// dot of the light direction and the direction from the light to the fragment ///
    float ldotd = dot(vec3(0.0, -1.0, 0.0), -l2fdir);
    
    /// attenuate the light strength based on the distance to the light and the light range ///
    float attenuation = (1.0 - min(l2f * WALL_SPOT_LIGHT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - WALL_SPOT_LIGHT_PARAMS[3].z) * WALL_SPOT_LIGHT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    float spottest = clamp(sign(WALL_SPOT_LIGHT_PARAMS[3].x - l2f) + sign(ldotd - WALL_SPOT_LIGHT_PARAMS[3].z), 0.0, 1.0) * sign(segment);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * WALL_SPOT_LIGHT_COLOR_PARAMS.xyz * attenuation, spottest);

    vec3 _halfv = normalize(l2fdir + viewDir);
    float ldoth = max(0.0, dot(l2fdir, _halfv));
    specTerm += mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * WALL_SPOT_LIGHT_COLOR_PARAMS.w, spottest);

    /// perform lighting with 1 more light toward point 1 ///
    /// determine the position of the light toward point 2 of the wall ///
    lightPos = lightPos - wallDirection * WALL_SPOT_LIGHT_PARAMS[0].z;

    /// direction of ray going from the fragment to the light position
    l2fdir = (lightPos - v_position);

    /// distance between fragment and light ///
    l2f = length(l2fdir);

    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);

    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    ndotl = max(0.0, dot(normalWS, l2fdir));

    /// dot of the light direction and the direction from the light to the fragment ///
    ldotd = dot(vec3(0.0, -1.0, 0.0), -l2fdir);
    
    /// attenuate the light strength based on the distance to the light and the light range ///
    attenuation = (1.0 - min(l2f * WALL_SPOT_LIGHT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - WALL_SPOT_LIGHT_PARAMS[3].z) * WALL_SPOT_LIGHT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    spottest = clamp(sign(WALL_SPOT_LIGHT_PARAMS[3].x - l2f) + sign(ldotd - WALL_SPOT_LIGHT_PARAMS[3].z), 0.0, 1.0) * clamp(sign(segment - 1.0), 0.0, 1.0);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * WALL_SPOT_LIGHT_COLOR_PARAMS.xyz * attenuation, spottest);

    _halfv = normalize(l2fdir + viewDir);
    ldoth = max(0.0, dot(l2fdir, _halfv));
    specTerm += mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * WALL_SPOT_LIGHT_COLOR_PARAMS.w, spottest);

    /// determine the position of the light toward point 2 of the wall ///
    lightPos = wallStartPoint + wallDirection * WALL_SPOT_LIGHT_PARAMS[0].z * (segment + 1.0) + vec3(WALL_SPOT_LIGHT_PARAMS[0].x, 0.0, WALL_SPOT_LIGHT_PARAMS[0].y)*WALL_SPOT_LIGHT_PARAMS[2].w;
    lightPos.y = WALL_SPOT_LIGHT_PARAMS[1].w;

    /// direction of ray going from the fragment to the light position
    l2fdir = (lightPos - v_position);

    /// distance between fragment and light ///
    l2f = length(l2fdir);

    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);

    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    ndotl = max(0.0, dot(normalWS, l2fdir));

    /// dot of the light direction and the direction from the light to the fragment ///
    ldotd = dot(vec3(0.0, -1.0, 0.0), -l2fdir);
    
    /// attenuate the light strength based on the distance to the light and the light range ///
    attenuation = (1.0 - min(l2f * WALL_SPOT_LIGHT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - WALL_SPOT_LIGHT_PARAMS[3].z) * WALL_SPOT_LIGHT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    spottest = clamp(sign(WALL_SPOT_LIGHT_PARAMS[3].x - l2f) + sign(ldotd - WALL_SPOT_LIGHT_PARAMS[3].z), 0.0, 1.0) * sign(WALL_SPOT_LIGHT_PARAMS[1].z - segment);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * WALL_SPOT_LIGHT_COLOR_PARAMS.xyz * attenuation, spottest);

    _halfv = normalize(l2fdir + viewDir);
    ldoth = max(0.0, dot(l2fdir, _halfv));
    specTerm += mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * WALL_SPOT_LIGHT_COLOR_PARAMS.w, spottest);
    
    /// perform lighting with 1 more light toward point 2 ///
    lightPos = lightPos + wallDirection * WALL_SPOT_LIGHT_PARAMS[0].z;

    /// direction of ray going from the fragment to the light position
    l2fdir = (lightPos - v_position);

    /// distance between fragment and light ///
    l2f = length(l2fdir);

    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);

    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    ndotl = max(0.0, dot(normalWS, l2fdir));

    /// dot of the light direction and the direction from the light to the fragment ///
    ldotd = dot(vec3(0.0, -1.0, 0.0), -l2fdir);
    
    /// attenuate the light strength based on the distance to the light and the light range ///
    attenuation = (1.0 - min(l2f * WALL_SPOT_LIGHT_PARAMS[3].y, 1.0));
    attenuation *= attenuation * pow(attenuation, 0.6);
    attenuation = clamp(clamp((ldotd - WALL_SPOT_LIGHT_PARAMS[3].z) * WALL_SPOT_LIGHT_PARAMS[3].w, 0.0, 1.0) * attenuation, 0.0, 1.0);
    spottest = clamp(sign(WALL_SPOT_LIGHT_PARAMS[3].x - l2f) + sign(ldotd - WALL_SPOT_LIGHT_PARAMS[3].z), 0.0, 1.0) * clamp(sign(WALL_SPOT_LIGHT_PARAMS[1].z - (segment + 1.0)), 0.0, 1.0);
    #ifdef EMISSION_LUM
    spotShadowTerm = occlusion;
    #endif
    
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * WALL_SPOT_LIGHT_COLOR_PARAMS.xyz * attenuation, spottest);
    diffuseTerm *= spotShadowTerm;

    _halfv = normalize(l2fdir + viewDir);
    ldoth = max(0.0, dot(l2fdir, _halfv));
    specTerm += mix(vec3(0.0, 0.0, 0.0), BRDF1_Unity_PBS(spotspeccolor, roughness, realRoughness, normalWS, _halfv, ldoth, ndotl) * attenuation * WALL_SPOT_LIGHT_COLOR_PARAMS.w, spottest);
    specTerm *= spotShadowTerm;

    #elif defined(WALL_POINT_LIGHTING)
    /// spec color is the same for every light ///
    vec3 spotspeccolor = specColor * WALL_POINT_LIGHT_COLOR_PARAMS.xyz;

    /// determine the number of lights, which is the wall width / light spacing ///
    ///float lightCount = WALL_SPOT_LIGHT_PARAMS[0].y;
    /// determine which section the fragment is in ...
    vec3 wallStartPoint = vec3(WALL_POINT_LIGHT_PARAMS[1].x, 0.0, WALL_POINT_LIGHT_PARAMS[1].y);
    vec3 wallDirection = normalize(vec3(WALL_POINT_LIGHT_PARAMS[2].x, 0.0, WALL_POINT_LIGHT_PARAMS[2].y) - wallStartPoint);
    
    /// need to determine the point along the wall corresponding to the vertex ///
    /// find distance from wallStartPoint to vertex (at the same height as wallStartPoint ///
    vec3 vertexWallDirection = (vec3(v_position.x, 0.0, v_position.z) - wallStartPoint);
    ///dot product of wallDirection and vertexWallDirection * length of vertexWallDirection ///
    float vertexWallDistance = dot(wallDirection, normalize(vertexWallDirection))*length(vertexWallDirection);
    float segment = floor(vertexWallDistance * WALL_POINT_LIGHT_PARAMS[0].w);
    
    /// Always 4 Lights ///
    /// determine the position of the light toward point 1 of the wall ///
    /// offset from the wall along the wall normal ///
    vec3 lightPos = wallStartPoint + wallDirection * WALL_POINT_LIGHT_PARAMS[0].z * segment + vec3(WALL_POINT_LIGHT_PARAMS[0].x, 0.0, WALL_POINT_LIGHT_PARAMS[0].y)*WALL_POINT_LIGHT_PARAMS[2].w;
    lightPos.y = 0.0;

    /// perform the light calculations ///
    /// direction of ray going from the fragment to the light position
    vec3 l2fdir = (lightPos - v_position);

    /// distance between fragment and light ///
    float l2f = length(l2fdir);

    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);

    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    float ndotl = max(0.0, dot(normalWS, l2fdir));
    
    /// attenuate the light strength based on the distance to the light and the light range ///
    float attenuation = (1.0 - min(l2f * WALL_POINT_LIGHT_PARAMS[3].y, 1.0));
    attenuation = clamp(attenuation * attenuation, 0.0, 1.0);
    float pointTest = clamp(sign(WALL_POINT_LIGHT_PARAMS[3].x - l2f), 0.0, 1.0) * sign(segment);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * WALL_POINT_LIGHT_COLOR_PARAMS.xyz * attenuation, pointTest);

    /// perform lighting with 1 more light toward point 1 ///
    /// determine the position of the light toward point 2 of the wall ///
    lightPos = lightPos - wallDirection * WALL_POINT_LIGHT_PARAMS[0].z;

    /// direction of ray going from the fragment to the light position
    l2fdir = (lightPos - v_position);

    /// distance between fragment and light ///
    l2f = length(l2fdir);

    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);

    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    ndotl = max(0.0, dot(normalWS, l2fdir));
    
    /// attenuate the light strength based on the distance to the light and the light range ///
    attenuation = (1.0 - min(l2f * WALL_POINT_LIGHT_PARAMS[3].y, 1.0));
    attenuation = clamp(attenuation * attenuation, 0.0, 1.0);
    pointTest = clamp(sign(WALL_POINT_LIGHT_PARAMS[3].x - l2f), 0.0, 1.0) * clamp(sign(segment - 1.0), 0.0, 1.0);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * WALL_POINT_LIGHT_COLOR_PARAMS.xyz * attenuation, pointTest);

    /// determine the position of the light toward point 2 of the wall ///
    lightPos = wallStartPoint + wallDirection * WALL_POINT_LIGHT_PARAMS[0].z * (segment + 1.0) + vec3(WALL_POINT_LIGHT_PARAMS[0].x, 0.0, WALL_POINT_LIGHT_PARAMS[0].y)*WALL_POINT_LIGHT_PARAMS[2].w;
    lightPos.y = 0.0;

    /// direction of ray going from the fragment to the light position
    l2fdir = (lightPos - v_position);

    /// distance between fragment and light ///
    l2f = length(l2fdir);

    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);

    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    ndotl = max(0.0, dot(normalWS, l2fdir));
    
    /// attenuate the light strength based on the distance to the light and the light range ///
    attenuation = (1.0 - min(l2f * WALL_POINT_LIGHT_PARAMS[3].y, 1.0));
    attenuation = clamp(attenuation * attenuation, 0.0, 1.0);
    pointTest = clamp(sign(WALL_POINT_LIGHT_PARAMS[3].x - l2f), 0.0, 1.0) * sign(WALL_POINT_LIGHT_PARAMS[1].z - segment);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * WALL_POINT_LIGHT_COLOR_PARAMS.xyz * attenuation, pointTest);

    /// perform lighting with 1 more light toward point 2 ///
    lightPos = lightPos + wallDirection * WALL_POINT_LIGHT_PARAMS[0].z;

    /// direction of ray going from the fragment to the light position
    l2fdir = (lightPos - v_position);

    /// distance between fragment and light ///
    l2f = length(l2fdir);

    /// normalized direction from fragment to the light ///
    l2fdir = normalize(l2fdir);

    /// calculate n.l and n.v for specular ///
    /// dot of fragment normal and direction to the light, capped at 0 (90deg) ///
    ndotl = max(0.0, dot(normalWS, l2fdir));
    
    /// attenuate the light strength based on the distance to the light and the light range ///
    attenuation = (1.0 - min(l2f * WALL_POINT_LIGHT_PARAMS[3].y, 1.0));
    attenuation = clamp(attenuation * attenuation, 0.0, 1.0);
    pointTest = clamp(sign(WALL_POINT_LIGHT_PARAMS[3].x - l2f), 0.0, 1.0) * clamp(sign(WALL_POINT_LIGHT_PARAMS[1].z - (segment + 1.0)), 0.0, 1.0);
    diffuseTerm += mix(vec3(0.0, 0.0, 0.0), ndotl * WALL_POINT_LIGHT_COLOR_PARAMS.xyz * attenuation, pointTest);

    #endif /// POINT_LIGHTING ///
    
   
    /// contribute ambient and diffuse lighting ///
    vec3 amCol = mix(AMBIENTCOLORMID, AMBIENTCOLORHIGH, clamp(dot(vec3(0.0,1.0,0.0), normalWS), 0.0, 1.0));
    albedo.rgb *= (mix(amCol, AMBIENTCOLORLOW, clamp(dot(vec3(0.0, -1.0, 0.0), normalWS), 0.0, 1.0)) + (diffuseTerm));

    #ifndef EMISSION_LUM
    //we do not want to apply these values here if we are using a lightmap, purely because this step should come after the lightmap application
        albedo.rgb += specTerm;
        //removing occlusion, it shouldnt influcence indirect
        //adjusted the roughness contribution here, it was previously too harsh so now its a minimum of 25% indirect. This is a magic number and the real solution is to fix the cubemap sampling issues
        albedo.rgb += CalculateIndirect(roughness, realRoughness, normalWS, oneMinusReflectivity, specColor, v_position, clamp(dot(normalWS, viewDir), 0.0, 1.0), viewDir, _ReflectionProbeStrength);
    #endif

    #endif /// defined(DIRECTIONAL_LIGHTING) || defined(SPOT_LIGHTING) || defined(WALL_SPOT_LIGHTING) ///

    #ifdef EMISSION_LUM
        
        vec3 white = vec3(1.0,1.0,1.0);

        //occlusion contribution is physically incorrect, but having it as a strength controllable artistic choice is preferable
        //1.0 metals should not use the lightmap sample since it can be black. Rather they should simply use white to be unchanged.
        float lightmap_multiplier = 1.5;
        float burn_factor = 0.05;
        
        vec3 lightmap_data = mix(emission * lightmap_multiplier, white, metallic);
        //Linear burn blend
        vec3 burned_albedo = ((albedo.rgb + (lightmap_data * burn_factor)) - burn_factor);

        //mask specular by ao
        specTerm *= occlusion;
        vec3 indirect = CalculateIndirect(roughness, realRoughness, normalWS, oneMinusReflectivity, specColor, v_position, clamp(dot(normalWS, viewDir), 0.0, 1.0), viewDir, _ReflectionProbeStrength);

        float indirect_occlusion = mix(1.0, occlusion, clamp(indirectMaskValues.y, 0.0, 1.0));
        vec3 maskedIndirect = indirect * indirect_occlusion;

        vec3 toneMappedColor = (burned_albedo * occlusion) + specTerm + maskedIndirect;
        vec3 maskedColor;
        if(_ToneMappingFunction == 1) {
            toneMappedColor = filmicToneMapping(toneMappedColor);
            maskedColor = LinearToGammaSpace(toneMappedColor);
        } else if(_ToneMappingFunction == 2) {
            toneMappedColor = linearToneMapping(toneMappedColor);
            maskedColor = LinearToGammaSpace(toneMappedColor);
        } else if(_ToneMappingFunction == 3) {
            toneMappedColor = simpleReinhardToneMapping(toneMappedColor);
            maskedColor = LinearToGammaSpace(toneMappedColor);
        } else if(_ToneMappingFunction == 4) {
            toneMappedColor = lumaBasedReinhardToneMapping(toneMappedColor);
            maskedColor = LinearToGammaSpace(toneMappedColor);
        } else if(_ToneMappingFunction == 5) {
            toneMappedColor = whitePreservingLumaBasedReinhardToneMapping(toneMappedColor);
            maskedColor = LinearToGammaSpace(toneMappedColor);
        } else if(_ToneMappingFunction == 6) {
            toneMappedColor = RomBinDaHouseToneMapping(toneMappedColor);
            maskedColor = LinearToGammaSpace(toneMappedColor);
        } else if(_ToneMappingFunction == 7) {
            toneMappedColor = Uncharted2ToneMapping(toneMappedColor);
            maskedColor = LinearToGammaSpace(toneMappedColor);
        } else if(_ToneMappingFunction == 8) {
            // special case for BHF to match what they expect to see. 
            // but it is physically wrong.
            toneMappedColor = (burned_albedo) + specTerm + maskedIndirect;
            maskedColor = LinearToGammaSpace(toneMappedColor) * occlusion;
        } else {
             maskedColor = LinearToGammaSpace(toneMappedColor);
        }

        //vec3 finalColor = maskedColor + maskedIndirect;
        //gl_FragColor = vec4(1,0,0,1);
        gl_FragColor = vec4(clamp(maskedColor, vec3(0.0,0.0,0.0), vec3(1.0,1.0,1.0)), outputAlpha);

    #else
        gl_FragColor = vec4(clamp(LinearToGammaSpace((albedo.rgb) * occlusion + emission), vec3(0.0,0.0,0.0), vec3(1.0,1.0,1.0)), outputAlpha);
    #endif
    

    
}
